#ifndef __24XUSB_H__
#define __24XUSB_H__

#include "24x.h"

//*************************************************************************
// USB definitons
//*************************************************************************

/* Power Management Register */
#define DISABLE_SUSPEND          0x00   
#define ENABLE_SUSPEND           0x01
#define SUSPEND_MODE             0x02
#define MCU_RESUME               0x04
#define EP0_FLUSH                0x40   
#define MCU_RESET                0x80

/* In & Out MAXP Register */
#define FIFO_SIZE_0              0x00  /* 0x00 * 8 = 0  */
#define FIFO_SIZE_8              0x01  /* 0x01 * 8 = 8  */
#define FIFO_SIZE_16             0x02  /* 0x02 * 8 = 16 */
#define FIFO_SIZE_64             0x08  /* 0x08 * 8 = 64 */

/* ENDPOINT0 CSR (Control Status Register) : Mapped to IN CSR1 */
#define EP0_OUT_PKT_READY        0x01  /* USB sets, MCU clears by setting SERVICED_OUT_PKT_RDY */
#define EP0_IN_PKT_READY         0x02  /* MCU sets, USB clears after sending FIFO */
#define EP0_SENT_STALL           0x04  /* USB sets */       
#define EP0_DATA_END             0x08  /* MCU sets */
#define EP0_SETUP_END            0x10  /* USB sets, MCU clears by setting SERVICED_SETUP_END */
#define EP0_SEND_STALL           0x20  /* MCU sets */
#define EP0_SERVICED_OUT_PKT_RDY 0x40  /* MCU writes 1 to clear OUT_PKT_READY */
#define EP0_SERVICED_SETUP_END   0x80  /* MCU writes 1 to clear SETUP_END */

#define EP0_WR_BITS              0xc0  

/* USB INT_REG & INT_MASK_REG */
#define EP0_INT                  0x01  // Endpoint 0, Control
#define EP1_INT                  0x02  // Endpoint 1, Bulk-In
#define EP2_INT                  0x04  // Endpoint 2, Bulk-In
#define EP3_INT                  0x08  // Endpoint 3, Bulk-Out
#define EP4_INT                  0x10  // Endpoint 4, Bulk-Out
#define SUSPEND_INT              0x20  
#define RESUME_INT               0x40  
#define RESET_INT                0x80  
#define ERROR_BC_INT             0x100  
#define ERROR_B_INT              0x100  
#define ERROR_C_INT              0x200  

/* EP1_IN_CSR & EP2_IN_CSR */
#define EPI_IN_PKT_READY         0x01  
#define EPI_FIFO_FLUSH           0x04
#define EPI_SEND_STALL           0x08  
#define EPI_SENT_STALL           0x10  
#define EPI_CDT                  0x20   
#define EPI_DMA_IPR_INT_MASK     0x40
#define EPI_AUTO_SET             0x80

#define EPI_WR_BITS              0x25 //0x4 is preferred  

/* EP3_OUT_CSR & EP4_OUT_CSR */
#define EPO_OUT_PKT_READY        0x01  
#define EPO_FIFO_FLUSH           0x04
#define EPO_SEND_STALL           0x08  
#define EPO_SENT_STALL           0x10  
#define EPO_CDT                  0x20   
#define EPO_DMA_OPR_INT_MASK     0x40
#define EPO_AUTO_CLR             0x80

#define EPO_WR_BITS              0x2c //0x4 is preferred

#endif //__24XUSB_H__
