#include "defines.h"
#include "gpstdlib.h" //required header for the I/O system
#include "gp32.h"
#include "smc.h"
#include "debug.h"
#include "gpmain.h"
#include "gpfont.h"

#define 	GPOS_RESID_FIO	2 
#define	LCD_WIDTH		320
#define	LCD_HEIGHT		240
#define	DEB_TIME		800000
#define	IDLE_TIME		10000


/* Global variables */
char g_string[256];
static GPDRAWSURFACE gpDraw;
extern volatile struct debug_struct debug_buffer[400];
static int debug_pos = 0;
extern unsigned char ss_buffer[16384+16];
extern unsigned long ss_currentblock;
unsigned long main_loop; // we don't initialise this on purpose. (used for not very random things)
unsigned int yyy;
unsigned long int debounce = DEB_TIME+1;
unsigned long int idlecounter = 0;
char message_code;

/* Protos - External */

extern void USBReset (void);
extern void USBComms (void); 
extern void USBInit (unsigned long numblocks, unsigned long blocksize);
extern void swi_install_irq( int num, void (*irq)(void) ); 
extern void cache_empty (void);
extern void ReadBlock (void);

/* Protos - Internal */

void ARMEnableINT(void);
void ARMDisableINT(void);
void Checkexit (void);
void USBISR(void) __attribute__ ((naked));
void Init(void);
void Puttxt (char *string, int posx, int posy, int color);
void Puttxtc (char *string, int posy, int color);
void GpSetPaletteEntry ( u8 i, u8 r, u8 g, u8 b );
void Puttxtc (char *string, int posy, int color);
void Puttxtb (char *string, int posy, int color);
void FillStatusWindow (char color);
void DrawDesktop (void);
void DrawStatusWindow (void);
char InitSMC (void);
char *DecodeDebug (int debug_pos);

/*****************************************************************************************/
/* MAIN PROGRAM **************************************************************************/
/*****************************************************************************************/

void GpMain (void * arg)
{
	char *str;
	Init();


	while(1)
	{
		if (debug_buffer[debug_pos].inUseFlag)
		{
			str = DecodeDebug (debug_pos++);			
			FillStatusWindow (78);
			Puttxtc (str, 130, 0xff);
			if (debug_pos > 399) debug_pos = 0;
			if (debounce>DEB_TIME) Puttxtc("When IDLE press SELECT to exit", 190, 0x01);		
			debounce = 0;
	 		Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0xFF);
		}
		if (debounce<DEB_TIME) 
		{
			debounce ++; 
			idlecounter = 0;
		}
		else 
		{
			if (debounce == DEB_TIME) 
			{
				debounce ++;
				FillStatusWindow (12);
				Puttxtc ("Idle", 130, 0xfe);
				Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0x01);
				Puttxtc("When IDLE press SELECT to exit", 190, 0xfe);
			}

				idlecounter ++;
				if (idlecounter == IDLE_TIME)
				{
					ARMDisableINT();
					idlecounter++;
 					FillStatusWindow (160);
					Puttxtc ("Saving Data", 130, 0xfe);
					Puttxtc("When IDLE press SELECT to exit", 190, 0x01);
					Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0xFF);
					cache_empty();
					FillStatusWindow (12);
					Puttxtc ("Idle", 130, 0xfe);
					Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0x01);
					Puttxtc("When IDLE press SELECT to exit", 190, 0xfe);
					ARMEnableINT();
				}

			Checkexit();
		}
	}
}

/*****************************************************************************************/
/* MAIN INITALIZATION ROUNTINE ***********************************************************/
/*****************************************************************************************/

void Init (void)
{
	/* USB */

	ARMDisableINT();
	rINTMSK=0xffffffff;
	rINTMSK |= iBIT_USBD; 
	USBReset();
	swi_install_irq (25, USBISR); 
	ClearPending (iBIT_USBD);
	ARMEnableINT();

	/* FAT */

	GpFatInit();

	/* Initialize Clock */

	GpClockSpeedChange(132000000,  0x3a011, 3); 

	/* Structures */

	memset (&debug_buffer, 0, sizeof (debug_buffer));

	/* RTC */

	rRTCCON |= 1;

	/* Initialize graphics */

	DrawDesktop();
	DrawStatusWindow ();

	/* Say hello */

	Puttxtb (">>>> phoDrive 0.3 (11-July-05) <<<<", 10, 0xff);
	Puttxtc ("By phoda (based on Squidge's GPDrive)", 30, 0xfe);


	/* SMC */

	if (InitSMC())
	{

 		Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0xFF);
	}
	else
	{
		Puttxtb("SMC unknown or not inserted!", 70, 0xfe); 
		DrawStatusWindow ();
		FillStatusWindow (160);
		Puttxtc("Session halted", 130, 0xfe);
		Puttxtb("Press SELECT to exit", 190, 0xff);
		while (1) Checkexit();
	}

}


/*****************************************************************************************/
/* SMC INITALIZATION ROUNTINE ************************************************************/
/*****************************************************************************************/

char InitSMC (void)
{
	unsigned char buf[32];
	unsigned char str[80];
	xsmc_init();
	xsmc_read_id (buf, 4);
	xsmc_detect (buf[0], buf[1], buf[2]);
	sprintf (str, "SMC Type: ");
	if (buf[0] == 0xec)
		strcat (str, "Samsung - Size: ");
	else if (buf[0] == 0x98)
		strcat (str, "Toshiba - Size: ");
	else
		return 0;

	switch (buf[1])
	{
		case 0x79:	strcat (str, "128 Mb");
						xsmc_make_l2p();
						USBInit (256000, 512);
						break;
						
		case 0x76:	strcat (str, "64 Mb");
						xsmc_make_l2p();
						USBInit (128000, 512);
						break;

		case 0x75:	strcat (str, "32 Mb");
						xsmc_make_l2p();
						USBInit (64000, 512);
						break;

		case 0x73:	strcat (str, "16 Mb");
						xsmc_make_l2p();
						USBInit (32000, 512);
						break;
		
		default:		return 0;
	} 
	Puttxtc (str, 70, 0xfe);
	return 1;
}


/*****************************************************************************************/
/* TEXT AND GRAPHICS ROUTINES ************************************************************/
/*****************************************************************************************/

void GpSetPaletteEntry ( u8 i, u8 r, u8 g, u8 b )  /* Sets a single GP32 palette entry */
{
	GP_PALETTEENTRY entry = (((((r>>3))&0x1f)<<11)|((((g>>3))&0x1f)<<6)|((((b>>3))&0x1f)<<1));
	GpPaletteEntryChange ( i, 1, &entry, 0 );
}


void Puttxtc (char *string, int posy, int color)
{
	unsigned int stringwidth, screenposition ;
	gp_str_func.strcpy(g_string, string);
	stringwidth = GpTextWidthGet(g_string);
	screenposition = (LCD_WIDTH - stringwidth)/2;
	GpTextOut(NULL, &gpDraw, screenposition, posy, g_string, color);
}

void Puttxtb (char *string, int posy, int color)
{
	unsigned int stringwidth, screenposition ;
	gp_str_func.strcpy(g_string, string);
	stringwidth = GpTextWidthGet(g_string);
	screenposition = (LCD_WIDTH - stringwidth)/2;
	GpTextOut(NULL, &gpDraw, screenposition, posy, g_string, color);
	GpTextOut(NULL, &gpDraw, screenposition+1, posy, g_string, color);
}

void DrawDesktop (void)
{
	GpLcdSurfaceGet(&gpDraw, 0);  
	GpSurfaceSet(&gpDraw);
	GpRectFill(NULL, &gpDraw, 0, 0, gpDraw.buf_w, gpDraw.buf_h, 0x01);
	yyy=0;
      Puttxtb ("---------------------------------------", 0, 0xfe);
      Puttxtb ("---------------------------------------", 1, 0xfe);
	do
	{
		Puttxtb ("|                                     |", yyy, 0xfe);
		yyy = yyy+2;
	}
	while (yyy < 225);
	Puttxtb ("---------------------------------------", 223, 0xfe);
	Puttxtb ("---------------------------------------", 224, 0xfe);
	Puttxtb ("---------------------------------------", 40, 0xfe);
	Puttxtb ("---------------------------------------", 41, 0xfe);
	GpRectFill(NULL, &gpDraw, 10, 9, (gpDraw.buf_w - 19), 38, 78);
}

void DrawStatusWindow (void)
{
	      Puttxtb ("STATUS", 100, 0xfe);
		yyy=110;
	      Puttxtb ("----------------------------------", 110, 0xfe);
	      Puttxtb ("----------------------------------", 111, 0xfe);
		do
		{
		Puttxtb ("|                                |", yyy, 0xfe);
		yyy = yyy+2;
		}
		while (yyy < 152);
		Puttxtb ("----------------------------------", 150, 0xfe);
		Puttxtb ("----------------------------------", 151, 0xfe);
}		

void FillStatusWindow (char color)
{
		GpRectFill(NULL, &gpDraw, 30, 119, 261, 38, color);
}

/*****************************************************************************************/
/* EXIT SEQUENCE *************************************************************************/
/*****************************************************************************************/

void Checkexit (void)
{
 	

	static unsigned char ignore_start = 0;
	
	if ((GpKeyGet() & GPC_VK_SELECT) == GPC_VK_SELECT)
	{
		ARMDisableINT();
		Puttxtb("Press SELECT to exit", 190, 0x01);
		Puttxtc(">>> DO NOT TURN THE GP32 OFF!!! <<<", 190, 0x01);
		Puttxtc("When IDLE press SELECT to exit", 190, 0x01);		
		FillStatusWindow (160);
		Puttxtc ("Saving Data", 130, 0xfe);
		cache_empty();
		FillStatusWindow (12);
		Puttxtc ("DONE!", 130, 0xff);
		Puttxtc ("Please release the SELECT button", 190, 0xff);
		do {} while ((GpKeyGet() & GPC_VK_SELECT) == GPC_VK_SELECT);
		ARMEnableINT();
		GpAppExit();
	}
}
 

/*****************************************************************************************/
/* ISR ***********************************************************************************/
/*****************************************************************************************/

void USBISR (void)
{
   asm volatile ("STMDB    r13!,{r0-r12,lr}\n");
	main_loop ++; // todo: fix, should really be called usbloop now?
	USBComms(); 
	ClearPending(iBIT_USBD);
   asm volatile ("LDMIA    r13!,{r0-r12,lr}\n"
                 "SUBS PC,lr,#4\n");
}
 

/*****************************************************************************************/
/* DEBUG CODES ***************************************************************************/
/*****************************************************************************************/

char *DecodeDebug (int debug_pos)
{
	static char str[256];
	switch (debug_buffer[debug_pos].flags1)
	{
		case 0: strcpy (str, "Empty"); break;
		case 1:  strcpy  (str, "Received address"); break;
		case 2:  sprintf (str, "Received config (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 3:  sprintf (str, "Interface set ok (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 4:  strcpy  (str, "Request to set invalid interface"); break;
		case 5:  strcpy  (str, "Unknown class request"); break;
		case 6:  sprintf (str, "Request descriptor (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 9:  sprintf (str, "Get config (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 10: sprintf (str, "Get interface (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 11: sprintf (str, "Get endpoint status (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 12: sprintf (str, "Clear EP feature (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 13: sprintf (str, "Set EP feature (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 14: sprintf (str, "Class request (%02X%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 15: sprintf (str, "Unrecognised command (%02X)", debug_buffer[debug_pos].v1); break;
		case 16: sprintf (str, "Write - Sector:%d Length:%d", (debug_buffer[debug_pos].v1 << 8)+debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 17: sprintf (str, "SCSI Status: %02X (dr:%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 18: sprintf (str, "SCSI Cmd: %02X (dtl:%02X%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 19: sprintf (str, "EP3 in (sl:%02X%02X m:%02X)", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 20: sprintf (str, "Read - Sector:%d Length:%d", (debug_buffer[debug_pos].v1 << 8)+debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 21: sprintf (str, "Read Capacity"); break;
		case 22: sprintf (str, "Read Format Capacity"); break;
		case 23: sprintf (str, "> SMC WRITE ERROR @ %02X%02X%02X <", debug_buffer[debug_pos].v1,debug_buffer[debug_pos].v2, debug_buffer[debug_pos].v3); break;
		case 24: sprintf (str, "> STOP! Writing has been disabled <"); break; 
		default: strcpy  (str, "Unknown response from USB interface"); break;
	}
	debug_buffer[debug_pos].inUseFlag = 0;
	return str;
}

/*****************************************************************************************/
/* ARM Interrupts (from Mr. Spiv) ********************************************************/
/*****************************************************************************************/

void ARMDisableINT( void ) 
{
asm volatile(""
 "mrs r0,CPSR   \n"
 "orr r0,r0,#192  \n"
 "msr CPSR,r0 \n"
 :
 :
 : "r0");
}

void ARMEnableINT( void ) 
{
asm volatile(""
 "mrs r0,CPSR   \n"
 "bic r0,r0,#128  \n"
 "msr CPSR,r0 \n"
 :
 :
 : "r0");
}


