#ifndef __SMC_H__
#define __SMC_H__

#include "defines.h"

//------------------------------------------------------------------------------
#define SMC_SAMSUNG_ID		0xEC
#define SMC_TOSHIBA_ID		0x98

#define SMC_16MB	0x73
#define SMC_32MB	0x75
#define SMC_64MB	0x76
#define SMC_128MB	0x79

#define LBA_UNUSED		0x80000000
#define LBA_RESERVED		0x80000001

#define BLOCK_UNUSED		0x80000000
#define BLOCK_RESERVED		0x80000001

/* SmartMedia Command */
#define SEQ_DATA_INPUT_CMD	0x80
#define READ1_CMD			0x00
#define READ1_1_CMD			0x01
#define READ2_CMD			0x50
#define READ_ID_CMD			0x90
#define RESET_CMD			0xFF
#define PAGE_PROGRAM_CMD	0x10
#define BLOCK_ERASE_CMD		0x60
#define BLOCK_ERASE_CFM_CMD	0xD0
#define READ_STATUS_CMD		0x70
#define RESET_PTR_CMD		0x00

//------------------------------------------------------------------------------
struct SMC_INFO
{
  u8  id[3];
  u32 bytes_per_page;
  u32 pages_per_block;
  u32 blocks;
  u32 mb;
};

//------------------------------------------------------------------------------
void xsmc_read_spare( u32 page, u8* buf, u8 length);
static u8 xsmc_write_page_ext( u32 page, u8 *buf, u8 *spare );
static u8 xsmc_create_new_block (u32 pblock, u32 lblock, u8 *data);
u8 xsmc_erase_block( u32 block);
int sm_ECCEncode(const u8 * p_buf, u8 * p_ecc);

// Externally accessible functions
u8 xsmc_read_block (u32 block, u8* buf, u32 mask);
void xsmc_read_id( u8* buf, u32 length);
u8 xsmc_write_block (u32 block, u8 *data);
void xsmc_init( void);
void xsmc_detect( u8 id1, u8 id2, u8 id3);
void xsmc_make_l2p( void);

extern u32 smc_l2p[0x2000];
extern u32 smc_p2l[0x2000];
extern struct SMC_INFO smc_info;

#endif
