//
// usb-interface
//
// 99% my own code. Sure, it looks messy at times, but pretty code is no use if it stops working afterwards :p

// BUGS & NOTES:
//
// Need to randomise block placement again. 
// We managed to get 380kb/sec when only erasing, so maybe put multiplane stuff back in?

#define DEBUG

#include "defines.h"
#include "gpgraphic.h"
#include "gpfont.h"
#include "gpstdio.h"
#include "gp32.h"
#include "usb100.h" 

#include "usb.h"
#include "smc.h"
#include "debug.h"

extern USB_DEVICE_DESCRIPTOR usb_descriptor;
extern USB_CONFIGURATION_DESCRIPTOR usb_config;
extern USB_INTERFACE_DESCRIPTOR usb_interface;
extern USB_ENDPOINT_DESCRIPTOR usb_ep1_tx;
extern USB_ENDPOINT_DESCRIPTOR usb_ep1_rx;

static unsigned int bulk_rx_buffer[1024];
static unsigned int bulk_rx_len = 0;

static unsigned char bulk_tx_buffer[1024];
static unsigned int bulk_tx_pos = 0;
static unsigned int bulk_tx_len = 0; // if bulk_tx_len is not zero, then packets will be created and sent out from pos bulk_tx_pos onwards until it's zero 
	
static unsigned char control_tx_buffer[256];
static unsigned int control_tx_pos = 0;
static unsigned int control_tx_len = 0;

static char medium_present = 1;
static char write_protected = 0;
char ep1_stalled = 0;
char ep3_stalled = 0;
#define DEVICE_DEFAULT 0
#define DEVICE_ADDRESSED 1
#define DEVICE_CONFIGURED 2
char DeviceState = DEVICE_DEFAULT;
unsigned char half_usb = 0;


unsigned long current_num_blocks = 0;
unsigned long current_block_size = 0;

#define Sense_NoMediumPresent 0
#define Sense_NoSense 1
#define Sense_IllegalRequest 2
#define Sense_WriteProtect 3

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

REQUEST_SENSE_DATA CurrentSense;

volatile struct debug_struct debug_buffer[400];

// starts from 0, when reached 399, we wrap around. 
static unsigned int debug_pos = 0, debug_rec_no = 0;

// general buffer for debug output/etc
static char str[256];
	
#define CFG_ACK_PACKET rEP0_CSR = uCSR0_SERVICED_OUT_PKT_RDY
#define CFG_STALL_PACKET rEP0_CSR = uCSR0_SERVICED_OUT_PKT_RDY | uCSR0_SEND_STALL
#define CFG_NO_DATA_TO_SEND rEP0_CSR = uCSR0_DATA_END

unsigned long ss_currentsector = 0;
unsigned long ss_currentblock = 0;
unsigned long ss_sectorsleft = 0;
unsigned long ss_offset = 0;
unsigned char ss_buffer[16384+16]; // 32 sectors plus room for spare area at end
unsigned long ss_offsetinsector = 0;
unsigned char ss_is_dirty = 0;
unsigned long ss_dirty_block = 0;   // dirty block number
unsigned char ss_writing_data = 0;

/* Cache */

void cache_write (u32 lblock, u8 *data);
char cache_read (u32 lblock, u8 *data);
u32 cache_id_clear (void);
u32 cache_id_get (void);
void cache_empty (void);

#define MAX_CACHE 64
/* Flags field bitmap */
#define CF_DIRTY 1
#define CF_USED 2

struct _gpdcache
{
	u32 EnteredTime;
	u32 LogicalBlock;
	u8 Flags; // see CF_
	u8 pad0; // 32-bit align
	u8 pad1;
	u8 pad2;
	u8 buffer[16384];
} CacheEntry [MAX_CACHE];

u32 cache_id = 0;

/******************************************************************************************************************************************************/

void ReadBlock (void)
{
	if (ss_is_dirty) // if we have a dirty block, write it to disk before reading a new one in
	{
		ss_is_dirty = 0;
		cache_write (ss_dirty_block, ss_buffer); 
	}
	if (ss_writing_data) // we are in the middle of accepting and writing data -= Must not be the first block =-
	{
		// if we are writing data to the smc, only read the old contents if we have to (eg. not filling a whole block)
		if (ss_sectorsleft < 32) 
		{
			unsigned long rmask;
			
			// We need to put "ss_sectorsleft" number of 0 bits into the LSB side of the reading mask
			// eg. one sector left to write, so mask should be 0xFFFFFFFE, two sectors left would be 0xFFFFFFFC
			// << This code assumes the data to be written is at the start of the block >>
			if (cache_read(ss_currentblock, ss_buffer) == 0)
			{
				rmask = (0xFFFFFFFF << ss_sectorsleft);
				xsmc_read_block (ss_currentblock, ss_buffer, rmask);
			}
		}
	}
	else
	{
		if (cache_read (ss_currentblock, ss_buffer) == 0) 
			xsmc_read_block (ss_currentblock, ss_buffer, 0xFFFFFFFF);
	}
	ss_dirty_block = ss_currentblock;
}

// Used by write10 command once at the command acceptance. #define to make it easier to change.
#define ReadBlock_Mask(mask) xsmc_read_block (ss_currentblock, ss_buffer, mask)

/******************************************************************************************************************************************************/

// called once at startup
void USBInit (unsigned long numblocks, unsigned long blocksize)
{
	u32 i;
	current_num_blocks = numblocks;
	current_block_size = blocksize;
	
	for (i = 0; i < MAX_CACHE; i ++)
		CacheEntry[i].Flags = 0;
		
	// initialise the sector buffer and get data ready for sector #0
	ReadBlock(); 
}

/******************************************************************************************************************************************************/

// Called when USB is reset //
void USBReset (void)
{
	rFUNC_ADDR_REG = uADDR_UPDATE | 0x00;	// set address to 0x00
	rPWR_REG = uPWR_MCU_RESET;					// software reset
	rPWR_REG = uPWR_EP0_FLUSH;					// empty ep0 fifo
	rPWR_REG = 0x00;								// normal op
	rINT_MASK_REG = ((~(0x80 + 8 + 2 + 1)) & 0x3FF);
	rINT_REG = 0x1FF;

	rEP0_MAXP = uMAXP_16;						// 16 bytes maximum input on endpoint 0
	rEP0_CSR = uCSR0_SERVICED_SETUP_END | uCSR0_SERVICED_OUT_PKT_RDY;
	rEP0_CSR = 0x00;

	rEP1_IN_CSR = uCSR1_FIFO_FLUSH | uCSR1_CLR_DATA_TOGGLE;
	rEP3_OUT_CSR = uCSR3_FIFO_FLUSH | uCSR3_CLR_DATA_TOGGLE;

	rEP1_IN_MAXP = uMAXP_64;					// 64 bytes input on endpoint 1
	rEP3_OUT_MAXP = uMAXP_64;					// 64 bytes output on endpoint 3

	bulk_tx_len = 0;
	bulk_rx_len = 0;
	control_tx_len = 0;
	half_usb = 0; 

	MassStorageInit();

	rEP1_IN_CSR = 0x00;
	rEP3_OUT_CSR = 0x00;
}

/******************************************************************************************************************************************************/

static void SendControlPacket (void)
{
	int i = 0;
	int maxbytes;
	
	maxbytes = 8;
	if (maxbytes > control_tx_len)
		maxbytes = control_tx_len;

	while (control_tx_len)
	{
		if (rINT_REG & 0x80) break; 
		rEP0_FIFO = control_tx_buffer[control_tx_pos];
		control_tx_pos ++;
		control_tx_len --;
		i ++;
		if (i >= maxbytes) break;
	}

	if ((rINT_REG & 0x80) == 0)
	{ 
		// if there is still more to go, then just send the packet, otherwise signify end of packet too.
		if (maxbytes == 8) 
		{
			rEP0_CSR = uCSR0_IN_PKT_RDY;              // more data to go
		}
		else if (maxbytes)
		{
			rEP0_CSR = uCSR0_IN_PKT_RDY | uCSR0_DATA_END;  // less than pipe size, so set data_end flag
			control_tx_len = 0;
		}
		else
		{
			rEP0_CSR = uCSR0_IN_PKT_RDY | uCSR0_DATA_END;  // no data whatsoever, so send zero length packet
			control_tx_len = 0;
		}
	}
}

/******************************************************************************************************************************************************/

void FullDebug (unsigned char v, unsigned char v1, unsigned char v2, unsigned char v3)
{
	debug_buffer[debug_pos].flags1 = v;
	debug_buffer[debug_pos].inUseFlag = 1;
	debug_buffer[debug_pos].v1 = v1;
	debug_buffer[debug_pos].v2 = v2;
	debug_buffer[debug_pos].v3 = v3;
	debug_pos ++;
	if (debug_pos > 399) debug_pos = 0;
}

/******************************************************************************************************************************************************/

void SimpleDebug (unsigned char v)
{
	FullDebug (v, 0, 0, 0);
}

/******************************************************************************************************************************************************/

// sends a packet out on ep0, assumes control_tx_buffer already filled in. Any packet currently being sent is truncated !
// This routine must be called for all valid ep0 requests, even if no data is to be transferred. Do not call in the case of a request error.
static void QueueControlReply (unsigned char length)
{
	CFG_ACK_PACKET;
	control_tx_pos = 0;
	control_tx_len = length;
	if (!length) 
		CFG_NO_DATA_TO_SEND;
}

/******************************************************************************************************************************************************/

static void ClearStall (char endpoint)
{
	endpoint &= 0x0F;
	if (endpoint == 3)
	{
		rEP3_OUT_CSR &= ~(uCSR3_SENT_STALL | uCSR3_SEND_STALL);
		rEP3_OUT_CSR |= uCSR3_CLR_DATA_TOGGLE;
		ep3_stalled = 0;
	}
	else if (endpoint == 1)
	{
		rEP1_IN_CSR &= ~(uCSR1_SENT_STALL | uCSR1_SEND_STALL);
		rEP1_IN_CSR |= uCSR1_CLR_DATA_TOGGLE;
		ep1_stalled = 0;
	}
}

/******************************************************************************************************************************************************/

static void SetStall (char endpoint)
{
	endpoint &= 0x0F;

	if (endpoint == 3)
	{
		rEP3_OUT_CSR |= uCSR3_SEND_STALL;
		ep3_stalled = 1;
	}
	else if (endpoint == 1)
	{
		ep1_stalled = 1;
		rEP1_IN_CSR |= uCSR1_SEND_STALL;
	}
}

/******************************************************************************************************************************************************/

// returns 1 for success, 0 for failure
static char ProcessDeviceRequests (usbrequest * usbreq)
{
	switch (usbreq->Request)
	{
		case USB_REQUEST_GET_STATUS:
			
			if (DeviceState == DEVICE_DEFAULT) return 0;
			if (usbreq->Value) return 0;
			if (usbreq->Index) return 0;
			if (usbreq->Length != 2) return 0;
			
			control_tx_buffer[0] = USB_GETSTATUS_SELF_POWERED;
			control_tx_buffer[1] = 0x00; // reserved
			QueueControlReply (2);
			return 1;

		case USB_REQUEST_GET_DESCRIPTOR:
		
			FullDebug (6, usbreq->ValueParts[0], usbreq->ValueParts[1], usbreq->Length);

			// We don't support string descriptors, so ensure the index is zero
			if (usbreq->Index) return 0;

			// We only support a single device and configuration descriptor, so ensure the descriptor index is also zero
			if (usbreq->ValueParts[0]) return 0;
		
			switch (usbreq->ValueParts[1]) // msb contains descriptor type, lsb contains index
			{
				case USB_DEVICE_DESCRIPTOR_TYPE: 
					if (usbreq->Length > 18) usbreq->Length = 18;
					memcpy (control_tx_buffer, &usb_descriptor, usbreq->Length);
					QueueControlReply (usbreq->Length);
					return 1;

				case USB_CONFIGURATION_DESCRIPTOR_TYPE:
					memcpy (control_tx_buffer, &usb_config, sizeof (usb_config)); control_tx_len = sizeof (usb_config);
					memcpy (&control_tx_buffer[control_tx_len], &usb_interface, sizeof (usb_interface)); control_tx_len += sizeof (usb_interface);
					memcpy (&control_tx_buffer[control_tx_len], &usb_ep1_rx, sizeof (usb_ep1_rx)); control_tx_len += sizeof (usb_ep1_rx);
					memcpy (&control_tx_buffer[control_tx_len], &usb_ep1_tx, sizeof (usb_ep1_tx)); control_tx_len += sizeof (usb_ep1_tx);
					if (control_tx_len > usbreq->Length)
						control_tx_len = usbreq->Length;
					QueueControlReply (control_tx_len);
					return 1;
					
				default:
					return 0;
			}
			
		case USB_REQUEST_SET_ADDRESS:
			
			FullDebug (1, usbreq->Value, DeviceState, usbreq->Index);

			// Both Index and Length MUST be zero
			if (usbreq->Index) return 0;
			if (usbreq->Length) return 0;

			// Complain if device address is > 127
			if (usbreq->Value > 127) return 0;
							
			rFUNC_ADDR_REG = uADDR_UPDATE + (usbreq->Value & 0x7F);		
			if (usbreq->Value) 
				DeviceState = DEVICE_ADDRESSED;
			else
				DeviceState = DEVICE_DEFAULT;
			QueueControlReply (0);
			return 1;

		case USB_REQUEST_SET_CONFIGURATION:
		
			FullDebug (2, usbreq->Value, usbreq->Index, usbreq->Length);

			if (DeviceState == DEVICE_DEFAULT) return 0;
			
			// Both Index and Length MUST be zero
			if (usbreq->Index) return 0;
			if (usbreq->Length) return 0;

			if (usbreq->Value == 0)
				DeviceState = DEVICE_ADDRESSED;
			else if (usbreq->Value == 1)
				DeviceState = DEVICE_CONFIGURED;
			else
				return 0; // invalid configuration value

			QueueControlReply (0);
			return 1;

		case USB_REQUEST_GET_CONFIGURATION:

			FullDebug (9, DeviceState, usbreq->Index, usbreq->Length);

			if (DeviceState == DEVICE_DEFAULT) return 0;
			
			if (usbreq->Index) return 0;
			if (usbreq->Length != 1) return 0;

			if (DeviceState == DEVICE_ADDRESSED)
				control_tx_buffer[0] = 0;
			else 
				control_tx_buffer[0] = 1;
			
			QueueControlReply (1);	
			return 1;
			
		default:
			// Unknown device request
			return 0;
	}			
}

/******************************************************************************************************************************************************/

static char ProcessInterfaceRequests (usbrequest * usbreq)
{
	switch (usbreq->Request)
	{
		case USB_REQUEST_GET_STATUS:

			if (DeviceState == DEVICE_DEFAULT) return 0;
			if (usbreq->Value) return 0;
			if (usbreq->Index) return 0; // we only support one interface
			if (usbreq->Length != 2) return 0;
			
			control_tx_buffer[0] = 0x00; // reserved
			control_tx_buffer[1] = 0x00; // reserved
			QueueControlReply (2);
			return 1;

		case USB_REQUEST_SET_INTERFACE:
		
			FullDebug (3, DeviceState, usbreq->Index, usbreq->Value);
			if (DeviceState < DEVICE_CONFIGURED) return 0;
			if (usbreq->Length) return 0;
			if (usbreq->Index) return 0; // we only support one interface
			if (usbreq->Value) return 0; // no alt setting allowed
			QueueControlReply (0);
			return 1;
						
		case USB_REQUEST_GET_INTERFACE:
		
			FullDebug (10, DeviceState, usbreq->Index, usbreq->Length);
			if (DeviceState < DEVICE_CONFIGURED) return 0;
			if (usbreq->Value) return 0;
			if (usbreq->Length != 1) return 0;
			if (usbreq->Index) return 0; // we only support one interface
			
			control_tx_buffer[0] = 0x00;
			QueueControlReply (1);
			return 1;
						
		default: 
		
			return 0;		
	}	
}

/******************************************************************************************************************************************************/

static char ProcessEndpointRequests (usbrequest * usbreq)
{
	switch (usbreq->Request)
	{
		case USB_REQUEST_GET_STATUS:
		
			FullDebug (11, DeviceState, usbreq->Index, usbreq->Length);
			if (usbreq->Value) return 0;
			if (usbreq->Length != 2) return 0;
			if (DeviceState == DEVICE_DEFAULT) return 0;
			if ((DeviceState == DEVICE_ADDRESSED) && (usbreq->Index)) return 0;
	
			switch (usbreq->Index)
			{
				case 0x00:

					control_tx_buffer[0] = 0x00;
					break;
					
				case 0x81:

					control_tx_buffer[0] = ep1_stalled;
					break;

				case 0x03:

					control_tx_buffer[0] = ep3_stalled;
					break;
				
				default:
					return 0;	
			}
			 
			control_tx_buffer[1] = 0x00;
			QueueControlReply (2);
			return 1;
		
		case USB_REQUEST_CLEAR_FEATURE:
		
			FullDebug (12, DeviceState, usbreq->Index, usbreq->Value);
			if (DeviceState == DEVICE_DEFAULT) return 0;
			if (usbreq->Length) return 0;
			if ((DeviceState == DEVICE_ADDRESSED) && (usbreq->Index)) return 0; // only ep0 available in address state
			if (usbreq->Value != USB_FEATURE_ENDPOINT_STALL) return 0; // only this feature available
			
			switch (usbreq->Index)
			{
				case 0x00: 

					break;				
					
				case 0x81:

					// If we stalled the command because the mass storage command was bad or unrecognised, then we need to send status now.
					// Note that calling ClearStall will clear the ep1_stalled flag, so we need to check the flag first
					if ((usbs.status == 1) && (ep1_stalled))
					{
						ClearStall (1);
						SendStatus();
						usbs.status = 0;
					}
					else
						ClearStall (1); // Just a normal clear stall request
					break;
					
				case 0x03:

					ClearStall (3);
					break;

				default:
					return 0;
			}			

			QueueControlReply (0);
			return 1;
							
		case USB_REQUEST_SET_FEATURE:

			FullDebug (13, DeviceState, usbreq->Index, usbreq->Value);
			if (DeviceState == DEVICE_DEFAULT) return 0;
			if (usbreq->Length) return 0;
			if ((DeviceState == DEVICE_ADDRESSED) && (usbreq->Index != 0)) return 0; // only ep0 available in address state
			if (usbreq->Value != USB_FEATURE_ENDPOINT_STALL) return 0; // only this feature available
			
			// Do not allow SetFeature to be used on ep0
			switch (usbreq->Index)
			{
				case 0x81:

					SetStall (1);
					break;
					
				case 0x03:

					SetStall (3);
					break;

				default:
					return 0;
			}			

		QueueControlReply (0);
		return 1;

		case USB_REQUEST_SYNC_FRAME:

				// FALL THROUGH //

		default:
		
			return 0;
	}
}

/******************************************************************************************************************************************************/

static char ProcessStandardRequests (usbrequest * usbreq)
{
	switch (usbreq->RequestType & 0x1F)
	{
		case RECIPIENT_DEVICE:
		
			return (ProcessDeviceRequests(usbreq));
		
		case RECIPIENT_INTERFACE:

			return (ProcessInterfaceRequests(usbreq));

		case RECIPIENT_ENDPOINT:
		
			return (ProcessEndpointRequests(usbreq));
			
		default:
		
			return 0;
	}
}			

/******************************************************************************************************************************************************/

static char ProcessClassRequests (usbrequest * usbreq)
{
	FullDebug (14, usbreq->Request, usbreq->Index, usbreq->Length);
	if ((usbreq->RequestType & 0x1F) != RECIPIENT_INTERFACE) return 0; 
	if (usbreq->Index) return 0; // we only support one interface

	// The below commands all expect a Value argument or zero, so put the check here. Will need to be moved if a command later on requires
	// a different value.
	if (usbreq->Value) return 0;
	
	switch (usbreq->Request)
	{
		case MASS_STORAGE_RESET:
			
			if (usbreq->Length) return 0;
			QueueControlReply (0);
			return 1;
			
		case MASS_STORAGE_GET_MAX_LUN:
		
			if (usbreq->Length != 1) return 0;
			control_tx_buffer[0] = 0x00;
			QueueControlReply (1);
			return 1;
			
		default:
		
			return 0;
	}
}

/******************************************************************************************************************************************************/

static void ProcessPkt (unsigned char *rx_buffer, unsigned char rx_len)
{
	usbrequest *usbreq = (usbrequest *)rx_buffer;
	
	switch ((usbreq->RequestType & 0x60) >> 5)
	{
		case TYPE_STANDARD:
					if (!ProcessStandardRequests (usbreq))
						CFG_STALL_PACKET;
					break;
		case TYPE_CLASS:
					if (!ProcessClassRequests (usbreq))
						CFG_STALL_PACKET;
					break;
		default: // all other
			CFG_STALL_PACKET;
			break;
	}
}

/******************************************************************************************************************************************************/

void ep0_process(void)
{
	long ep0csrval = rEP0_CSR;
	
	if (ep0csrval & uCSR0_SETUP_END)
		USBReset();

	if (ep0csrval & uCSR0_SENT_STALL)
		USBReset();
	
	if (ep0csrval & uCSR0_OUT_PKT_RDY)
	{
		long bytestorcv = rEP0_OUT_CNT;
	
		// pkt ready
		if (bytestorcv)
		{
			int bufpos = 0;
			unsigned char control_rx_buffer[16];
			
			while (bytestorcv--)
				control_rx_buffer[bufpos++] = (unsigned char)((rEP0_FIFO) & 0xFF);					

			ProcessPkt (control_rx_buffer, bufpos);
		}
	}
	
	if (((ep0csrval & uCSR0_IN_PKT_RDY) == 0) && (control_tx_len))
		SendControlPacket();
}

/******************************************************************************************************************************************************/

void USBComms(void)  
{
	long usbintval;
		
	usbintval = rINT_REG;
	if (!usbintval) return;
	
	if (usbintval & uINT_RESET)
	{
		USBReset();
		rINT_REG = uINT_RESET; // RESET_INT should be cleared after ResetUsbd().      
		usbintval &= ~(uINT_RESET);
	}

	if (usbintval & uINT_EP0)
	{
		rINT_REG = uINT_EP0;
		usbintval &= ~(uINT_EP0);
		ep0_process();
	}

	if (usbintval & uINT_EP1)
	{
		rINT_REG = uINT_EP1;
		usbintval &= ~(uINT_EP1);
		ep1_process();
	}

	if (usbintval & uINT_EP3)
	{
		rINT_REG = uINT_EP3;
		usbintval &= ~(uINT_EP3);
		ep3_process();
	}

	// Any other interrupt just gets cleared.
	if (usbintval)
	{
		rINT_REG = usbintval;
	}
}

/********************************/
/* Mass storage device specific */
/*                              */
/* SCSI Stuff                   */
/*                              */
/********************************/

// Should really be in a seperate file, but no time //

/******************************************************************************************************************************************************/

void MassStorageInit (void)
{
	usbs.dwsignature = 0x53425355;
	usbs.status = 0;
	ss_writing_data = 0;
}

/******************************************************************************************************************************************************/

void SendBulk (int size)
{
	// if no outgoing data transfer in progress, start transmission of data. Assumes data will never be more than 64 bytes.
	// (sectors etc are sent seperately for speed reasons, and to make this routine easier)
	if (!bulk_tx_len)
	{
		bulk_tx_len = size;
		bulk_tx_pos = 0;
		
		while (bulk_tx_len)
		{
			rEP1_FIFO = bulk_tx_buffer[bulk_tx_pos];
			bulk_tx_pos ++;
			bulk_tx_len --;
		}
		
		if ((rINT_REG & 0x80) == 0) // if usb reset didn't occur, set the flag to actually let the data go
			rEP1_IN_CSR = (rEP1_IN_CSR & (~uCSR1_WR_BITS) | uCSR1_IN_PKT_RDY);
		else
			return;
	}			
}

/******************************************************************************************************************************************************/

void SendBulkData (int size)
{
	if (!bulk_tx_len)
	{
		SendBulk (size);
		// since we'll never send 64 bytes, queue the status packet for sending as soon as this one has finished sending
		// as we'll never queue anything but a status packet, this becomes really easy -> just set bulk_tx_len to a non-zero number :)
		bulk_tx_len = 13; // 13 just happens to be the size of the status packet, but can be anything > 0 as described above
	}
}

/******************************************************************************************************************************************************/

void CreateSenseData (char Sense)
{
	CurrentSense.ResponseCode = 0x70;
	CurrentSense.Valid = 0;
	CurrentSense.SegmentNum = 0;
	CurrentSense.Reserved0 = 0;
	CurrentSense.WrongLenIndicator = 0;
	CurrentSense.EndofMedium = 0;
	CurrentSense.FileMark = 0;
	CurrentSense.Info_0 = 0;
	CurrentSense.Info_1 = 0;
	CurrentSense.Info_2 = 0;
	CurrentSense.Info_3 = 0;
	CurrentSense.AdditionalSenseLen = 0x0A;
	CurrentSense.CommandSpecInfo_0 = 0;
	CurrentSense.CommandSpecInfo_1 = 0;
	CurrentSense.CommandSpecInfo_2 = 0;
	CurrentSense.CommandSpecInfo_3 = 0;
	CurrentSense.FieldReplacableUnitCode = 0;
	CurrentSense.SenseKeySpec_0 = 0;
	CurrentSense.SenseKeySpecValid = 0;
	CurrentSense.SenseKeySpec_1 = 0;
	CurrentSense.SenseKeySpec_2 = 0;
	 
	switch (Sense)
	{
		case Sense_NoMediumPresent:
				CurrentSense.SenseKey = 2;
				CurrentSense.ASC = 0x3a;
				CurrentSense.ASCQ = 0x00;
				break;
		case Sense_NoSense:
				CurrentSense.SenseKey = 0;
				CurrentSense.ASC = 0;
				CurrentSense.ASCQ = 0;
				break;
		case Sense_WriteProtect:
				CurrentSense.SenseKey = 7;
				CurrentSense.ASC = 0x27;
				CurrentSense.ASCQ = 0x00;
				break;
		case Sense_IllegalRequest:
		default: 
				CurrentSense.SenseKey = 5;
				CurrentSense.ASC = 0x24;
				CurrentSense.ASCQ = 0x00;
				break;
	}
}

/******************************************************************************************************************************************************/

// Send USBS-style response
static char SendStatus (void)
{
	int i;
	// signature already filled in by init code
	// tag filled in by USBC input packet
	// data residue filled in by whoever acted upon the packet
	// status should also be filled in by same

	// so, all thats left is that we queue the packet for sending.
	if ((usbc.uccb[0] == 0x28) || (usbc.uccb[0] == 0x1a)) 
		FullDebug (17, 0, 0, 0);
	for (i = 0; i < 13; i ++) // for some reason, memcpy doesn't work, so have to use this
		bulk_tx_buffer[i] = ((unsigned char *)&usbs)[i];
	SendBulk (13);	 	
}

/******************************************************************************************************************************************************/

// 
static void RequestSense (void)
{
	int i;
	unsigned char *p = (unsigned char *)&CurrentSense;

	for (i = 0; i < sizeof(REQUEST_SENSE_DATA); i ++)
	{
		bulk_tx_buffer[i] = *p;
		p++;
	}
}

/******************************************************************************************************************************************************/

// take the bulk_rx_buffer and process it, possible ending up sending a packet back or stalling the appropriate endpoint
// return 1 for ok, 0 to stall
static char ProcessSCSICmd (void)
{
	long dataxferlen;
	
	unsigned char SCSIInquiryData[44] = // this data ripped from another program and modified to suit to I know what flags to set :)
	{
		0x00, //= Device class
		0x80, //= Removable medium
		0x02, //= SPEC = SCSI2
		0x02, //= Data format = 2 (SCSI-2)
		0x00, //= Additional length
		0x00, 0x00, 0x00, //
		' ' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' , // = Manufacturer
		'g' , 'p' , 'D' , 'r' , 'i' , 'v' , 'e' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' , ' ' ,
		'0' , '0' , '0' , '5' , // = Revision
		0x30, 0x39, 0x2F, 0x32, 0x34, 0x2F, 0x39, 0x38, // = Vendor unique (chopped off)
	};
	
	switch (usbc.uccb[0])
	{

// should these be #define'd instead of hex values?

		case 0x12: // inquiry

			FullDebug (18, usbc.uccb[0], usbc.uccb[4], usbc.uccb[5]);
			if (usbc.uccb[1] || usbc.uccb[2]) // Logical unit, EVPD or page code > 0 specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}
			CreateSenseData(Sense_NoSense);
			dataxferlen = MIN(usbc.uccb[4], sizeof(SCSIInquiryData));
			memcpy (bulk_tx_buffer, SCSIInquiryData, dataxferlen);
			SendBulkData (dataxferlen);
			usbs.dwdataresidue -= dataxferlen;
			return 1;

		case 0x03: // request sense

			FullDebug (18, usbc.uccb[0], usbc.uccb[4], usbc.uccb[5]);
			if (usbc.uccb[1]) // Logical unit specified, but we can't modify the sense data, so just return fail
				return 0;
			RequestSense();
			dataxferlen = MIN(usbc.uccb[4], sizeof(REQUEST_SENSE_DATA));
			SendBulkData (dataxferlen);
			usbs.dwdataresidue -= dataxferlen;
			return 1;

		case 0x00: // test unit ready

			if (usbc.uccb[1]) // Logical unit specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}
			if (medium_present)
				CreateSenseData(Sense_NoSense);
			else
				CreateSenseData(Sense_NoMediumPresent);
			return medium_present;

		case 0x1e: // prevent/allow medium removal

			if ((usbc.uccb[1]) || (usbc.uccb[4] & 1)) // Logical unit or "prevent removal" specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}

			CreateSenseData(Sense_NoSense);
			return 1;

		case 0x28: // read 10

			if (usbc.uccb[1]) // Logical unit, DPO, FUA, or RelAdr specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}

			if (medium_present && (!ep1_stalled))
			{
				int i;
								
				usbs.dwdataresidue = 0;
				CreateSenseData(Sense_NoSense);
				ss_sectorsleft = ((usbc.uccb[7] << 8) + (usbc.uccb[8]));
				ss_currentsector = (usbc.uccb[2] << 24);
				ss_currentsector += (usbc.uccb[3] << 16);
				ss_currentsector += (usbc.uccb[4] << 8);
				ss_currentsector += (usbc.uccb[5]);				
				ss_offset = ((ss_currentsector & 0x1F) << 9); // get modulo 32 and multiply by 512
				ss_offsetinsector = 0;
				FullDebug (20, (ss_currentsector >> 8) & 0xFF, ss_currentsector & 0xFF, ss_sectorsleft);
				if (ss_currentblock != (ss_currentsector >> 5))
				{
					ss_currentblock = (ss_currentsector >> 5); // get block by dividing sector number by 32
					ReadBlock (); 
				}

				SendSectors(); // start off the transfer (64 bytes)
				return 1;				
			}
			else
			{
				CreateSenseData(Sense_NoMediumPresent);
				return 0;
			}

		case 0x25: // read capacity

			// check for logical unit, RelAdr, LBA, or PMI
			if ((usbc.uccb[1]) || (usbc.uccb[2]) || (usbc.uccb[3]) || (usbc.uccb[4]) || (usbc.uccb[5]) || (usbc.uccb[8])) 
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}
			FullDebug (21, 0,0,0);

			if (medium_present)
			{
				CreateSenseData(Sense_NoSense);
				bulk_tx_buffer[0] = 0x00; // We only support cards upto 8gb ;)
				bulk_tx_buffer[1] = ((current_num_blocks-1) >> 16) & 0xFF;
				bulk_tx_buffer[2] = ((current_num_blocks-1) >> 8) & 0xFF;
				bulk_tx_buffer[3] = (current_num_blocks-1) & 0xFF;
				bulk_tx_buffer[4] = 0x00; // and the same for block sizes...
				bulk_tx_buffer[5] = (current_block_size >> 16) & 0xFF;
				bulk_tx_buffer[6] = (current_block_size >> 8) & 0xFF;
				bulk_tx_buffer[7] = (current_block_size) & 0xFF;
				usbs.dwdataresidue -= 8;
				SendBulkData (8);
				return 1;
			}
			else
			{
				CreateSenseData(Sense_NoMediumPresent);
				return 0;
			}

		case 0x23: // read format capacities

			if (usbc.uccb[1]) // Logical unit specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}

			FullDebug (22, 0,0,0);
			if (medium_present)
			{
				dataxferlen = MIN(((usbc.uccb[7] << 8) | (usbc.uccb[8])), 12); 
				CreateSenseData(Sense_NoSense);
				
				bulk_tx_buffer[0] = 0x00; // format capacity header (number of bytes to follow - 8)
				bulk_tx_buffer[1] = 0x00;
				bulk_tx_buffer[2] = 0x00;
				bulk_tx_buffer[3] = 0x08;

				bulk_tx_buffer[4] = 0x00;
				bulk_tx_buffer[5] = ((current_num_blocks-1) >> 16) & 0xFF;
				bulk_tx_buffer[6] = ((current_num_blocks-1) >> 8) & 0xFF;
				bulk_tx_buffer[7] = (current_num_blocks-1) & 0xFF;

				bulk_tx_buffer[8] = 0x02; // descriptor code (2 = current capacity)

				bulk_tx_buffer[9] = (current_block_size >> 16) & 0xFF;
				bulk_tx_buffer[10] = (current_block_size >> 8) & 0xFF;
				bulk_tx_buffer[11] = (current_block_size) & 0xFF;
				
				usbs.dwdataresidue -= dataxferlen;
				SendBulkData (dataxferlen);
				return 1;
			}
			else
			{
				CreateSenseData(Sense_NoMediumPresent);
				return 0;
			}
	
		case 0x1b: // start stop unit

			FullDebug (18, usbc.uccb[0], usbc.uccb[4], usbc.uccb[5]);
			if (usbc.uccb[1] & 0xFE) // Logical unit specified
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}

			if (usbc.uccb[4] & 2) // test the "LoEj" (Eject) bit. Accept if not set.
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}
			else
			{
				CreateSenseData(Sense_NoSense);
				return 1;
			}

		case 0x1a: // mode sense 6 -- NEEDS CHECKING - Do we have the specs for this one? They all seem to conflict, so have to guess.

			FullDebug (18, usbc.uccb[0], usbc.uccb[4], usbc.uccb[5]);
			if (usbc.uccb[1]) // Logical unit or DBD
			{
				CreateSenseData(Sense_IllegalRequest); // "Disable block descriptor" set and we don't support it
				return 0;
			}
			if ((usbc.uccb[4] > 0x0c) || (usbc.uccb[2] != 0x3F)) // asking for too much info or a certain page -> default answer 
			{
				bulk_tx_buffer[0] = 0x1C;          // Taken from Firefly's Mass Storage driver
				bulk_tx_buffer[1] = 0x06;
				bulk_tx_buffer[2] = 0x00;
				bulk_tx_buffer[3] = 0x00;
				bulk_tx_buffer[4] = 0x00;
				bulk_tx_buffer[5] = 0x00;
				bulk_tx_buffer[6] = 0x00;
				bulk_tx_buffer[7] = 0x00;
				dataxferlen = MIN(usbc.uccb[4], 8);
				SendBulkData (dataxferlen);
				usbs.dwdataresidue -= dataxferlen;
				return 1;
			}
			// otherwise, if medium inserted, return number of blocks and block size, otherwise fail
			if (medium_present)
			{
				bulk_tx_buffer[0] = 0x3F;   // what my SMC reader seems to spit out...
				bulk_tx_buffer[1] = 0x00;
				if (write_protected)
					bulk_tx_buffer[2] = 0x80;
				else
					bulk_tx_buffer[2] = 0x00;
				bulk_tx_buffer[3] = 0x08;
				bulk_tx_buffer[4] = 0x00; // medium type (0 = default)
				bulk_tx_buffer[5] = (current_num_blocks >> 16) & 0xFF;
				bulk_tx_buffer[6] = (current_num_blocks >> 8) & 0xFF;	
				bulk_tx_buffer[7] = (current_num_blocks) & 0xFF;
				bulk_tx_buffer[8] = 0x00; // reserved
				bulk_tx_buffer[9] = (current_block_size >> 16) & 0xFF;
				bulk_tx_buffer[10] = (current_block_size >> 8) & 0xFF;
				bulk_tx_buffer[11] = (current_block_size) & 0xFF;
				dataxferlen = MIN(usbc.uccb[4], 12);
				SendBulkData (dataxferlen);
				usbs.dwdataresidue -= dataxferlen;
				return 1;
			}
			else
			{
				CreateSenseData (Sense_NoMediumPresent);			
				return 0;
			}

		case 0x2a: // write_10 - the biggee !

			if (usbc.uccb[1] || write_protected) // Logical unit, DPO, FUA, or RelAdr specified
			{
				CreateSenseData(Sense_IllegalRequest);
				if (write_protected)
					FullDebug (24, 0, 0, 0);
				return 0;
			}

			usbs.dwdataresidue = 0;
			CreateSenseData(Sense_NoSense);
			ss_sectorsleft = ((usbc.uccb[7] << 8) + (usbc.uccb[8]));
			ss_currentsector = (usbc.uccb[2] << 24);
			ss_currentsector += (usbc.uccb[3] << 16);
			ss_currentsector += (usbc.uccb[4] << 8);
			ss_currentsector += (usbc.uccb[5]);				
			ss_offset = (ss_currentsector & 0x1F) << 9;
			ss_offsetinsector = 0;
			ss_writing_data = 1;
			FullDebug (16, (ss_currentsector >> 8) & 0xFF, ss_currentsector & 0xFF, ss_sectorsleft);
			if (ss_currentblock != (ss_currentsector >> 5))
			{
				unsigned long rmask;
				unsigned long our_sectsleft;
				unsigned long ofsec = (ss_currentsector & 0x1F);

				ss_currentblock = (ss_currentsector >> 5);
				ss_dirty_block = ss_currentblock;
				if (cache_read (ss_currentblock, ss_buffer) == 0)
				{
					// Build a mask of the sectors to read, bit 0 = sector 0, bit 1 = sector 1, etc. 
					// 1 means read the sector, 0 means don't bother as it's going to be written over anyway					
						
					// find number of sectors to write in current block first
					if ((ofsec + ss_sectorsleft) > 32) our_sectsleft = 32 - ofsec; 
					else our_sectsleft = ss_sectorsleft;
						
					// now we got that, we can compute the mask
					rmask = 0xFFFFFFFF;
					rmask <<= (our_sectsleft + ofsec); // shift '0' bits into the sectors that are going to be written
					rmask |= ((1 << ofsec) - 1);       // and replace the bits on the right that are not going to be written		
					ReadBlock_Mask (rmask);
				}
			}
			else // we already have the block in memory, so don't bother reading it
				ss_dirty_block = ss_currentblock;
			return 1;

		case 0x2f: // verify (faked)

			FullDebug (18, usbc.uccb[0], usbc.uccb[4], usbc.uccb[5]);
			if (usbc.uccb[1]) // Logical address, RelAdr, ByteChk, DPO
			{
				CreateSenseData(Sense_IllegalRequest);
				return 0;
			}
			if (medium_present)
			{
				CreateSenseData(Sense_NoSense);
				return 1;
			}
			else
			{
				CreateSenseData(Sense_NoMediumPresent);
				return 0;
			}

		default: 
			FullDebug (15, usbc.uccb[0], 0, 0);
			return 0;	
	}
}

/******************************************************************************************************************************************************/

static void ReceiveWriteData (unsigned char write_len)
{
	unsigned char *fifodata = &ss_buffer[ss_offset + ss_offsetinsector];
	ss_is_dirty = 1;
	while (write_len >= 8)
	{
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		*fifodata++ = rEP3_FIFO;
		ss_offsetinsector += 8;
		write_len -= 8;
	}
	// get the packet in
	while (write_len)
	{
		*fifodata++ = rEP3_FIFO;
		ss_offsetinsector++;
		write_len --;
	}
	if (ss_offsetinsector >= 512)
	{
		ss_offsetinsector = 0;
		ss_offset += 512;
		ss_sectorsleft --;
		if (ss_sectorsleft)
		{
			if (ss_offset >= 16384)
			{
				ss_offset = 0;
				ss_currentblock ++;
				ReadBlock(); // write the current block, and read the next
			}
		}
		else
		{
			// no sectors left
			ReadBlock(); // flush the changes to the current block to disk
			ss_writing_data = 0;
			ss_offset = 0;
			ss_offsetinsector = 0;
			SendStatus(); // send out our status block
		}
	} // ss_offsetinsector >= 512
}

/******************************************************************************************************************************************************/

static char ValidCBW(void)
{
	if (bulk_rx_len < 15) return 0;              // not a whole command block wrapper
	if (bulk_rx_len < 15+usbc.uccblen) return 0;	 // control block has incorrect length

	if (usbc.dwsignature != 0x43425355) return 0; // "USBC" reversed (signature)
	
	return 1;
}

/******************************************************************************************************************************************************/

static void ReceiveUSBC (unsigned char rcv_len)
{
	if (rcv_len)
	{
		while (rcv_len)
		{
			((unsigned char *)&usbc)[bulk_rx_len++] = rEP3_FIFO;
			rcv_len --;			
		}
		if (ValidCBW())
		{
			usbs.dwtag = usbc.dwtag;
			usbs.dwdataresidue = usbc.dwdataxferlen; // may have to switch this between msb/lsb
			if (!ProcessSCSICmd())
			{
				// command failed. If there's a data phase, then stall the relevent pipe depending on direction (plus USBS failure status later), 
				// else just send out the failed USBS
				usbs.status = 1;
				// check for data phase
				if (usbc.dwdataxferlen)				
				{
					if (usbc.ucflags & 0x80)	// host expecting data from us
						SetStall (1);
					else								// host sending data to us
						SetStall (3);
				}
				else // no data phase, so just send out failed status
					SendStatus();
			}
			else // scsi command complete ok, send out success status if no data phase
			{
				usbs.status = 0;
				if ((!usbc.dwdataxferlen) && (!bulk_tx_len))
					SendStatus(); 
				else
					half_usb = 1; // usb operation not fully completed yet
			}
		}
		else
		{
			// CBW not valid, so the only choice we have is stalling both pipes as we don't know what its expecting
			SetStall (3);
			SetStall (1);
		}
		bulk_rx_len = 0;
	}
}

/******************************************************************************************************************************************************/

// general processing routine for when something happens to the ep3 endpoint (incoming data)
static void ep3_process (void)
{
	long ep3_csr = rEP3_OUT_CSR;
	
	if (ep3_csr & uCSR3_OUT_PKT_RDY)
	{
		long ep3length = rEP3_OUT_CNT & 0xFF;
		
		//FullDebug (19, (ss_sectorsleft >> 8) & 0xFF, (ss_sectorsleft & 0xFF), ep3length | (ss_writing_data << 7)); 
		if (ss_writing_data)
			ReceiveWriteData (ep3length);
		else
			ReceiveUSBC (ep3length);
		rEP3_OUT_CSR = (ep3_csr & (~uCSR3_WR_BITS) & (~uCSR3_OUT_PKT_RDY)); // modify: taken from Samsung/Firefly instead of just clearing OUT_PKT_RDY
	}
}

/******************************************************************************************************************************************************/

void SendSectors (void)
{
	int i = 0;
	unsigned char *fifodata = &ss_buffer[ss_offset + ss_offsetinsector];
	
	// load 64 bytes into the usb fifo in chunks of 8
		
	while (ss_offsetinsector < 512)
	{
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		rEP1_FIFO = *fifodata ++;
		ss_offsetinsector += 8;
		if ((ss_offsetinsector & 0x3F) == 0)       // 64 bytes sent yet?
			break;
	}

	// if the bus hasn't reset, send them out
	if ((rINT_REG & 0x80) == 0) // if usb reset didn't occur, set the flag to actually let the data go
	{
		rEP1_IN_CSR = (rEP1_IN_CSR & (~uCSR1_WR_BITS) | uCSR1_IN_PKT_RDY);
		// whilst they are sending, see if we need to read another block from disk...
		if (ss_offsetinsector >= 512)
		{
			ss_offsetinsector = 0;
			ss_sectorsleft --;
			if (ss_sectorsleft)
			{
				// more to go, same block or next one?
				ss_offset += 512; 
				if (ss_offset >= 16384)
				{
					// next block
					ss_offset = 0;
					ss_currentblock ++;
					ReadBlock ();
				}
			}
			else
				bulk_tx_len = 13; // send the status packet
		} // otherwise we have not sent a complete sector yet
	} 
}

/******************************************************************************************************************************************************/

// general processing routine for when something happens to the ep1 endpoint (outgoing data)
static void ep1_process (void)
{
	// transmission routines. status & stall checking, etc should go above this.
	if ((rEP1_IN_CSR & uCSR1_IN_PKT_RDY) == 0) 
	{
		if (bulk_tx_len)
		{
			// Our transmission buffer is now empty, need to send the status packet queued up from before
			bulk_tx_len = 0;
			SendStatus();
			half_usb = 0; // status sent - end of command
		}
		else if (ss_sectorsleft)
			SendSectors();		
	}	
}

/* Cache */

/*

Some unused functions.. Ho hum... Originally intended to be used for timed writes (cache emptying) when we had nothing else to do...

char rtc_getsec (void)
{
	unsigned char bcdsec = rBCDSEC;
	
	return (((bcdsec >> 4)*10)+(bcdsec & 0x0F));
}

u32 rtc_timestamp (void)
{
	unsigned char d, h,m,s;
	
	do
	{
		d = rBCDDAY;
		h = rBCDHOUR;
		m = rBCDMIN;
		s = rBCDSEC;
	} while (!s);
	
	return ((d << 24) | (h << 16) | (m << 8) | (s));
}
*/

// nb: if more than 4,294,967,295 blocks are written in one session, this routine may cause the wrong blocks to be expunged from the cache !
// (no data loss will occur however)
u32 cache_id_get (void)
{
	if (cache_id < 0xFFFFFFFF) cache_id ++;
	return cache_id;	
}

u32 cache_id_clear (void)
{
	cache_id = 0;
}

// find logical block in cache. if entry can not be found in the cache, returns 0. 
char cache_read (u32 lblock, u8 *data)
{
	u32 i;
	u32 pos;
	
	for (i = 0; i < MAX_CACHE; i ++)
	{
		if ((CacheEntry[i].Flags & (CF_USED)) && (CacheEntry[i].LogicalBlock == lblock))
		{
			for (pos = 0; pos < 16384; pos ++)
				data[pos] = CacheEntry[i].buffer[pos];
			return 1;
		}
	}
	return 0;
}

void cache_write (u32 lblock, u8 *data)
{
	u32 i;
	u32 FreeBlock = 0xFFFFFFFF;
	u32 Oldest = 0xFFFFFFFF;
	u32 OldestPos = 0xFFFFFFFF;
	u32 pos;
	// try to find in use structure with block same as the one we have been given. If we find, then increment lastaccess and return. 
	
	for (i = 0; i < MAX_CACHE; i ++)
	{
		if ((CacheEntry[i].Flags & (CF_USED | CF_DIRTY)) == (CF_USED | CF_DIRTY))
		{
			if (CacheEntry[i].LogicalBlock == lblock)
			{
				u32 pos;
				// found it, update last access, copy over the data, and we are done
				CacheEntry[i].EnteredTime = cache_id_get();
				for (pos = 0; pos < 16384; pos ++)
					CacheEntry[i].buffer[pos] = data[pos];
				return;
			} 
			// look for the oldest block whilst we are here just in case we need to throw one out
			if (CacheEntry[i].EnteredTime < Oldest)
			{
				Oldest = CacheEntry[i].EnteredTime;
				OldestPos = i;
			}		
		}
		// find a free block whilst we are here just in case we need one
		if ((CacheEntry[i].Flags & (CF_USED | CF_DIRTY)) == 0) 
			FreeBlock = i;
	}
	// Block not already present in the cache, so if we have at least one place free in the cache, dump the data there and return
	if (FreeBlock != 0xFFFFFFFF)
	{
		CacheEntry[FreeBlock].EnteredTime = cache_id_get();
		CacheEntry[FreeBlock].LogicalBlock = lblock;
		CacheEntry[FreeBlock].Flags = (CF_USED | CF_DIRTY);
		for (pos = 0; pos < 16384; pos ++)
			CacheEntry[FreeBlock].buffer[pos] = data[pos];
		return;
	}
	// Nothing free in the cache, will have to write something to make space.

	// If OldestPos not changed, we are in trouble
	if (OldestPos == 0xFFFFFFFF) return;
	// Write out...
	if (!xsmc_write_block (CacheEntry[OldestPos].LogicalBlock, CacheEntry[OldestPos].buffer))
	{
		write_protected = 1;
		FullDebug (23, CacheEntry[OldestPos].LogicalBlock >> 16, CacheEntry[OldestPos].LogicalBlock >> 8, CacheEntry[OldestPos].LogicalBlock);
	}
	// Now accept incoming
	CacheEntry[OldestPos].EnteredTime = cache_id_get();
	CacheEntry[OldestPos].LogicalBlock = lblock;
	CacheEntry[OldestPos].Flags = (CF_USED | CF_DIRTY);
	for (pos = 0; pos < 16384; pos ++)
		CacheEntry[OldestPos].buffer[pos] = data[pos];
	return;
}

void cache_empty (void)
{
	u32 i;
	
	for (i = 0; i < MAX_CACHE; i ++)
	{
		if ((CacheEntry[i].Flags & (CF_USED | CF_DIRTY)) == (CF_USED | CF_DIRTY))
		{
			if (!xsmc_write_block (CacheEntry[i].LogicalBlock, CacheEntry[i].buffer))
			{
				write_protected = 1;
				FullDebug (23, CacheEntry[i].LogicalBlock >> 16, CacheEntry[i].LogicalBlock >> 8, CacheEntry[i].LogicalBlock);
			}
			CacheEntry[i].Flags = 0;
		}	
	}
}


